/*
 * proto.c - PM VNC Viewer, RFB Protocol Handling
 */

#include <stdio.h>
#include <stdlib.h>

#define INCL_PM
#include <os2.h>

#define DEBUG

#include "pmvncdef.h"
#include "vncauth.h"

/*
 * Hold Remote Frame Buffer Info.
 */
 
static  rfbServerInitMsg    ServerInit ;
static  PUCHAR              ServerDesk ;

/*
 * Format & Encodings
 */
 
PVNCCTX     ctxPix = &ctxPix32 ;    /* Pixel format & drawing   */

#define MAX_ENCODINGS   10

CARD32          DefaultEncodingsTab[] = {
    rfbEncodingHextile,
    rfbEncodingCoRRE,
    rfbEncodingRRE,
    rfbEncodingRaw                  /* at last */
} ;

#define     DefaultEncodingsNum (sizeof(DefaultEncodingsTab)/sizeof(CARD32))

CARD32      LocalEncodingsTab[MAX_ENCODINGS] ;
CARD16      LocalEncodingsNum ;

/*
 * chkVersion - Exchange Protocol Version
 */

static  BOOL    authWillWork = TRUE ;

static  BOOL    chkVersion(void)
{
    rfbProtocolVersionMsg   pv ;
    int     major, minor ;
    
    if (netRecv(pv, sz_rfbProtocolVersionMsg) != TRUE) {
        netFail("failed to recv RFB Version") ;
        return FALSE ;
    }
    pv[sz_rfbProtocolVersionMsg] = '\0' ;

    if (sscanf(pv, rfbProtocolVersionFormat, &major, &minor) != 2) {
        netFail("failed to get Protocol Version") ;
	return FALSE ;
    }
    
    TRACE("Server RFB Protocol Version %d.%d\n", major, minor) ;
    
    if (major != rfbProtocolMajorVersion) {
        netFail("RFB Major Version Unmatch") ;
	return FALSE ;
    }
    if (minor < rfbProtocolMinorVersion) {
        authWillWork = FALSE ;
    } else {
        major = rfbProtocolMajorVersion ;
        minor = rfbProtocolMinorVersion ;
    }
    sprintf(pv, rfbProtocolVersionFormat, major, minor) ;
    
    if (netSend(pv, sz_rfbProtocolVersionMsg) != TRUE) {
        netFail("failed to send Protocol Version") ;
	return FALSE ;
    } 
    
    return TRUE ;
}

/*
 * chkAuth - Authentication
 */

static  UCHAR   authChallenge[CHALLENGESIZE] ;

static  BOOL    chkAuth(void)
{
    CARD32  authScheme, reasonLen, authResult ;
    PUCHAR  reason ;
    
    /*
     * get auth. scheme
     */
     
    if (netRecv((PUCHAR) &authScheme, sizeof(CARD32)) != TRUE) {
        netFail("failed to recv Auth. Scheme") ;
	return FALSE ;
    }
    authScheme = swap32(authScheme) ;
    
    TRACE("Auth. Scheme %d\n", authScheme) ;
    
    if (authScheme == rfbNoAuth) {
        TRACE("no authentication needed\n") ;
	return TRUE ;
    }
    if (authScheme == rfbConnFailed) {
        if (netRecv((PUCHAR) &reasonLen, sizeof(CARD32)) != TRUE) {
	    netFail("failed to recv reason length") ;
	    return FALSE ;
	}
	reasonLen = swap32(reasonLen) ;
	
	if ((reason = malloc(reasonLen)) == NULL) {
	    netFail("failed to alloc reason buffer") ;
	    return FALSE ;
	}
	if (netRecv(reason, reasonLen) != TRUE) {
	    netFail("failed to recv reason") ;
	    return FALSE ;
	}
	netFail(reason) ;       /* don't free 'reason'  */
	return FALSE ;
    }	
    if (authScheme != rfbVncAuth) {
        netFail("unknown authenticaion scheme") ;
        return FALSE ;
    }

    /*
     * case of rfbVncAuth
     */
     
    if (! authWillWork) {
        netFail("VNC Server uses old Auth. scheme") ;
        return FALSE ;
    }
    if (netRecv(authChallenge, CHALLENGESIZE) != TRUE) {
        netFail("failed to recv auth. challenge") ;
        return FALSE ;
    }
    netNotify(WM_VNC_AUTH, 0, authChallenge) ;
        /*
	 * then window thread encode 'authChallenge' 
	 * and calls protoSendAuth
	 */
    if (netRecv((PUCHAR) &authResult, sizeof(CARD32)) != TRUE) {
        netFail("failed to recv auth. result") ;
        return FALSE ;
    }
    authResult = swap32(authResult) ;
	
    TRACE("auth. result %d\n", authResult) ;
	
    switch (authResult) {
    case rfbVncAuthOK :
        TRACE("rfbVncAuthOK\n") ;
        return TRUE ;
    case rfbVncAuthFailed :
        netFail("rfbVncAuthFailed") ;
	return FALSE ;
    case rfbVncAuthTooMany :
        netFail("rfbVncAuthTooMany") ;
	return FALSE ;
    default :
        netFail("unknown VNC auth. result") ;
	return FALSE ;
    }
    return FALSE ;      /* not reach here   */
}

/*
 * protoSendAuth - Send Authentication Data
 */

BOOL    protoSendAuth(void)
{
    if (netSend(authChallenge, CHALLENGESIZE) != TRUE) {
        netFail("failed to Send Auth.") ;
	return FALSE ;
    }
    return TRUE ;
}

/*
 * chkInit - Client/Server Initialization Message
 */

static  BOOL    chkInit(void)
{
    rfbClientInitMsg    ci ;
    rfbServerInitMsg    si ;
    
    /*
     * send Client Initialization Message
     */
     
    ci.shared = (SessOptShared ? 1 : 0) ;

    if (netSend((PUCHAR) &ci, sz_rfbClientInitMsg) != TRUE) {
        netFail("failed to send ClientInit") ;
	return FALSE ;
    }
    
    /*
     * recv Server Initialization Message
     */

    if (netRecv((PUCHAR) &si, sz_rfbServerInitMsg) != TRUE) {
        netFail("failed to recv ServerInit") ;
	return FALSE ;
    }
    
    si.framebufferWidth  = swap16(si.framebufferWidth)  ;
    si.framebufferHeight = swap16(si.framebufferHeight) ;
    si.format.redMax   = swap16(si.format.redMax)   ;
    si.format.greenMax = swap16(si.format.greenMax) ;
    si.format.blueMax  = swap16(si.format.blueMax)  ;
    si.nameLength = swap32(si.nameLength) ;
    
    memcpy(&ServerInit, &si, sizeof(rfbServerInitMsg)) ;

    if ((ServerDesk = malloc(si.nameLength + 1)) == NULL) {
        netFail("failed to alloc desktop name") ;
	return FALSE ;
    }
    if (netRecv(ServerDesk, si.nameLength) != TRUE) {
        netFail("failed to recv desktop name") ;
	return FALSE ;
    }

#if 1
    printf("Remote Desktop [%s] %d x %d\n", 
        ServerDesk, si.framebufferWidth, si.framebufferHeight) ;
    printf("Server's natural format is\n") ;
    printf("    bitsPerPixel %d, depth %d, bigEndian %s, trueColour %s\n",
                si.format.bitsPerPixel, si.format.depth,
		(si.format.bigEndian ? "TRUE" : "FALSE"),
		(si.format.trueColour ? "TRUE" : "FALSE")) ;
    printf("    red   Max %d, shift %d\n", 
                si.format.redMax,   si.format.redShift) ;
    printf("    green Max %d, shift %d\n",
                si.format.greenMax, si.format.greenShift) ;
    printf("    blue  Max %d, shift %d\n", 
                si.format.blueMax,  si.format.blueShift) ;
#endif

    netNotify(WM_VNC_INIT, 0, NULL) ;
    
    return TRUE ;
}

/*
 * protoQueryDesktop - query size and name of remote desktop
 */

PUCHAR  protoQueryDesktop(PSIZEL siz)
{
    siz->cx = ServerInit.framebufferWidth  ;
    siz->cy = ServerInit.framebufferHeight ;
    return ServerDesk ;
}

/*
 * protoConnInit - Initialize RFB Connection
 */

BOOL    protoConnInit(void)
{
    if (chkVersion() != TRUE) {
        return FALSE ;
    }
    if (chkAuth() != TRUE) {
        return FALSE ;
    }
    if (chkInit() != TRUE) {
        return FALSE ;
    }
    return TRUE ;
}

/*
 * protoSendFmtEnc - send SetPixelFormat & SetEncodings
 */

static  BOOL    inputReady = FALSE ;
 
BOOL    protoSendFmtEnc(void)
{
    rfbSetPixelFormatMsg    spf ;
    rfbSetEncodingsMsg      *se ;
    CARD32  *encs  ;
    int     i, len ;
    UCHAR   buf[sz_rfbSetEncodingsMsg + sizeof(CARD32) * MAX_ENCODINGS] ;
    
    if (SessOptUse8bits) {
        ctxPix = &ctxPix8  ;
    } else {
        ctxPix = &ctxPix32 ;
    }
    
    /*
     * send SetPixelFormat Message
     */

    spf.type = rfbSetPixelFormat ;
    spf.format = ctxPix->rectFormat ;
    
    spf.format.redMax   = swap16(spf.format.redMax)   ;
    spf.format.greenMax = swap16(spf.format.greenMax) ;
    spf.format.blueMax  = swap16(spf.format.blueMax)  ;
    
    if (netSend((PUCHAR) &spf, sz_rfbSetPixelFormatMsg) != TRUE) {
        netFail("failed to send SetPixelFormat") ;
	return FALSE ;
    }
    
    /*
     * send SetEncodings Message
     */
     
    se   = (rfbSetEncodingsMsg *) &buf[0] ;
    encs = (CARD32 *) &buf[sz_rfbSetEncodingsMsg] ;
    
    se->type = rfbSetEncodings ;
    len = sz_rfbSetEncodingsMsg ;
    LocalEncodingsNum = 0 ;
    
    TRACE("Pref Encoding %d\n", SessOptEncoding) ;

    *encs++ = swap32(rfbEncodingCopyRect)  ;        /* place at top */
    LocalEncodingsNum += 1 ;
    len += sizeof(CARD32)  ;

    if (SessOptEncoding >= rfbEncodingRaw && 
                SessOptEncoding <= rfbEncodingHextile) {
        *encs++ = swap32(SessOptEncoding) ;         /* preferred    */
        LocalEncodingsNum += 1 ;
        len += sizeof(CARD32)  ;
    }

    for (i = 0 ; i < DefaultEncodingsNum ; i++) {   /* as default   */
        *encs++ = swap32(DefaultEncodingsTab[i]) ;
	LocalEncodingsNum += 1 ;
	len += sizeof(CARD32)  ;
    }
    se->nEncodings = swap16(LocalEncodingsNum) ;

    if (netSend(buf, len) != TRUE) {
        netFail("failed to send SetEncodings") ;
	return FALSE ;
    }
    
    inputReady = TRUE ;

    return TRUE ;
}

/*
 * protoSendRequest - send FramebufferUpdateRequest
 */

BOOL    protoSendRequest(BOOL inc, PRECTL rect)
{
    rfbFramebufferUpdateRequestMsg  fur ;
    RECTL   full ;
    
    fur.type = rfbFramebufferUpdateRequest ;
    fur.incremental = inc ? 1 : 0 ;
    
    if (rect == NULL) {
	full.xLeft   = 0 ;
	full.yBottom = 0 ;
	full.xRight  = ServerInit.framebufferWidth  ;
	full.yTop    = ServerInit.framebufferHeight ;
        rect = &full ;
    }
    fur.x = swap16(rect->xLeft) ;
    fur.w = swap16(rect->xRight - rect->xLeft) ;
    fur.y = swap16(ServerInit.framebufferHeight - rect->yTop) ;
    fur.h = swap16(rect->yTop - rect->yBottom) ;
    
    if (netSend((PUCHAR) &fur, sz_rfbFramebufferUpdateRequestMsg) != TRUE) {
        netFail("failed to send FramebufferUpdateRequest") ;
	return FALSE ;
    }
    return TRUE ;
}

/*
 * protoSetColourMapEntries
 *      not expected for TRUE COLOR, but read and skip message
 */

static  BOOL    protoSetColourMapEntries(rfbServerToClientMsg *msg)
{
    int     i ;
    CARD16  rgb[3] ;
    
    TRACE("rfbSetColourMapEntries - unexpected, skip\n") ;
    
    if (netRecv(((PUCHAR) msg) + 1, sz_rfbSetColourMapEntriesMsg - 1) != TRUE) {
        netFail("failed to recv rfbSetColourMapEntires") ;
	return FALSE ;
    }
    
    msg->scme.firstColour = swap16(msg->scme.firstColour) ;
    msg->scme.nColours    = swap16(msg->scme.nColours)    ;
    
    for (i = 0 ; i < msg->scme.nColours ; i++) {
        if (netRecv((PUCHAR) rgb, sizeof(rgb)) != TRUE) {
	    netFail("failed to recv rfbSetColourMapEntires, RGB") ;
	    return FALSE ;
	}
    }
    return TRUE ;
}

/*
 * protoBell - notify to window thread to ring a bell
 */

static  BOOL    protoBell(rfbServerToClientMsg *msg)
{
    TRACE("rfbBell\n") ;
    netNotify(WM_VNC_BELL, 0, NULL) ;
    return TRUE ;
}

/*
 * protoServerCutText - read cut text and notify to window
 */

static  BOOL    protoServerCutText(rfbServerToClientMsg *msg)
{
    PUCHAR  str ;
    
    TRACE("rfbServerCutText\n") ;
    
    if (netRecv(((PUCHAR) msg) + 1, sz_rfbServerCutTextMsg - 1) != TRUE) {
        netFail("failed to recv rfbServerCutTextMsg") ;
	return FALSE ;
    }

    msg->sct.length = swap32(msg->sct.length) ;

    if ((str = malloc(msg->sct.length)) == NULL) {
        netFail("failed to alloc for rfbServerCutText") ;
	return FALSE ;
    }
    
    if (netRecv(str, msg->sct.length) != TRUE) {
        netFail("failed to recv rfbServerCutText, text") ;
	return FALSE ;
    }
    
    netNotify(WM_VNC_CLIP, 0, str) ;    /* 'str' will disposed in window */

    return TRUE ;
}

/*
 * protoFramebufferUpdate
 */

static  BOOL    protoFramebufferUpdate(rfbServerToClientMsg *msg)
{
    rfbFramebufferUpdateRectHeader  rect ;
    int     i    ;
    BOOL    stat ;
    SIZEL   rfb  ;
    int     x, y, w, h ;
    
    protoQueryDesktop(&rfb) ;
    
    if (netRecv(((PUCHAR) msg) + 1, sz_rfbFramebufferUpdateMsg - 1) != TRUE) {
        netFail("failed to recv rfbFramebufferUpdateMsg") ;
	return FALSE ;
    }
    msg->fu.nRects = swap16(msg->fu.nRects) ;

    for (i = 0 ; i < msg->fu.nRects ; i++) {

        if (netRecv((PUCHAR) &rect, sz_rfbFramebufferUpdateRectHeader) != TRUE) {
	    netFail("failed to recv rfbFramebufferUpdateRectHeader") ;
	    return FALSE ;
	}
	rect.r.x = x = swap16(rect.r.x) ;
	rect.r.y = y = swap16(rect.r.y) ;
	rect.r.w = w = swap16(rect.r.w) ;
	rect.r.h = h = swap16(rect.r.h) ;
	rect.encoding = swap32(rect.encoding) ;
	
        if ((rect.r.x + rect.r.w) > ServerInit.framebufferWidth) {
	    netFail("rect too large (width)") ;
	    netDump((PUCHAR) &rect, sz_rfbFramebufferUpdateRectHeader) ;
	    return FALSE ;
	}
	if ((rect.r.y + rect.r.h) > ServerInit.framebufferHeight) {
	    netFail("rect too large (height)") ;
	    netDump((PUCHAR) &rect, sz_rfbFramebufferUpdateRectHeader) ;
	    return FALSE ;
	}
	if ((rect.r.w * rect.r.h) == 0) {
	    TRACE("empty rectangle, skip\n") ;
	    continue ;
	}
	
	switch (rect.encoding) {
        case rfbEncodingRaw      :
	    stat = (*ctxPix->rectRaw) (&rect.r) ;
	    break ;
	case rfbEncodingCopyRect :
	    stat = (*ctxPix->rectCopy) (&rect.r) ;
	    break ;
	case rfbEncodingRRE      :
	    stat = (*ctxPix->rectRRE) (&rect.r) ;
	    break ;
	case rfbEncodingCoRRE    :
	    stat = (*ctxPix->rectCoRRE) (&rect.r) ;
	    break ;
	case rfbEncodingHextile  :
	    stat = (*ctxPix->rectTile) (&rect.r) ;
	    break ;
	default :
	    netFail("unknown rect encoding") ;
	    stat = FALSE ;
            break ;
	}
	if (stat != TRUE) {
	    return FALSE ;
	}
	WinPostMsg(hwndClient, WM_VNC_UPDATE,
       	    MPFROM2SHORT(x, rfb.cy - y - h), MPFROM2SHORT(w, h)) ;
    }    
    return TRUE ;
}

/*
 * protoDispatch - entry for Server -> Client Message Handling
 */

BOOL    protoDispatch(void)
{
    rfbServerToClientMsg    msg ;
    
    if (netRecv((PUCHAR) &msg, 1) != TRUE) {
        netFail("failed to recv Server -> Client Msg") ;
	return FALSE ;
    }
    
    switch (msg.type) {
    case rfbSetColourMapEntries :
        return protoSetColourMapEntries(&msg) ;
    case rfbFramebufferUpdate :
        return protoFramebufferUpdate(&msg) ;
    case rfbBell :
        return protoBell(&msg) ;
    case rfbServerCutText :
        return protoServerCutText(&msg) ;
    }
    netFail("unknown message from Server") ;
    return FALSE ;
}

/*
 * protoSendMouEvent - send Pointer Event
 */

BOOL    protoSendMouEvent(USHORT state, PPOINTL pt)
{
    rfbPointerEventMsg  pe ;
    
    if (SessOptViewonly) {
        return TRUE ;
    }
    if (! inputReady) {
        return TRUE ;
    }
    
    pe.type = rfbPointerEvent ;
    pe.buttonMask = (UCHAR) (state & 0xff) ;
    pe.x = swap16((USHORT) pt->x) ;
    pe.y = swap16((USHORT) pt->y) ;
    
    return netSend((PUCHAR) &pe, sz_rfbPointerEventMsg) ;
}

/*
 * protoSendKeyEvent - send Keyboard Event
 */

BOOL    protoSendKeyEvent(BOOL down, ULONG key)
{
    rfbKeyEventMsg  ke ;
    
    if (SessOptViewonly) {
        return TRUE ;
    }
    if (! inputReady) {
        return TRUE ;
    }
    
    ke.type = rfbKeyEvent  ;
    ke.down = down ? 1 : 0 ;
    ke.key  = swap32(key)  ;
    
    return netSend((PUCHAR) &ke, sz_rfbKeyEventMsg) ;
}
